@@ -79,7 +79,8 @@ get_arpadate (char *d_string)
 	time_t now;
 
 	/* RFC822 format string borrowed from GNU shellutils date.c */
-	const char *format = "%a, %_d %b %Y %H:%M:%S %z";
+	/* Using %d instead of %_d, the second one isn't portable */
+	const char *format = "%a, %d %b %Y %H:%M:%S %z";
 
 	now = time(NULL);
 
