@@ -10,7 +10,7 @@
  See COPYRIGHT for the license
 
 */
-#define VERSION "2.63"
+#define VERSION "2.65"
 #define _GNU_SOURCE
 
 #include <sys/socket.h>
@@ -55,21 +55,21 @@ bool_t use_oldauth = False;		/* use old AUTH LOGIN username style */
 
 #define ARPADATE_LENGTH 32		/* Current date in RFC format */
 char arpadate[ARPADATE_LENGTH];
-char *auth_user = (char)NULL;
-char *auth_pass = (char)NULL;
-char *auth_method = (char)NULL;		/* Mechanism for SMTP authentication */
-char *mail_domain = (char)NULL;
-char *from = (char)NULL;		/* Use this as the From: address */
+char *auth_user = NULL;
+char *auth_pass = NULL;
+char *auth_method = NULL;		/* Mechanism for SMTP authentication */
+char *mail_domain = NULL;
+char *from = NULL;		/* Use this as the From: address */
 char *hostname;
 char *mailhost = "mailhub";
-char *minus_f = (char)NULL;
-char *minus_F = (char)NULL;
+char *minus_f = NULL;
+char *minus_F = NULL;
 char *gecos;
-char *prog = (char)NULL;
+char *prog = NULL;
 char *root = NULL;
 char *tls_cert = "/etc/ssl/certs/ssmtp.pem";	/* Default Certificate */
-char *uad = (char)NULL;
-char *config_file = (char)NULL;		/* alternate configuration file */
+char *uad = NULL;
+char *config_file = NULL;		/* alternate configuration file */
 
 headers_t headers, *ht;
 
@@ -97,6 +97,20 @@ static char hextab[]="0123456789abcdef";
 
 ssize_t outbytes;
 
+#if defined (__SVR4) && defined (__sun)
+/*
+strndup() - Unimplemented by the Solaris libc.
+*/
+char * strndup(char const *s, size_t n)
+{
+	size_t len = strnlen(s, n);
+	char *new = malloc(len + 1);
+	if(new == NULL) return NULL;
+	new[len] = '\0';
+	return memcpy(new, s, len);
+}
+#endif
+
 /*
 log_event() -- Write event to syslog (or log file if defined)
 */
@@ -261,7 +275,7 @@ char *strip_post_ws(char *str)
 
 	p = (str + strlen(str));
 	while(isspace(*--p)) {
-		*p = (char)NULL;
+		*p = 0;
 	}
 
 	return(p);
@@ -287,7 +301,7 @@ char *addr_parse(char *str)
 		q++;
 
 		if((p = strchr(q, '>'))) {
-			*p = (char)NULL;
+			*p = 0;
 		}
 
 #if 0
@@ -310,7 +324,7 @@ char *addr_parse(char *str)
 	q = strip_post_ws(p);
 	if(*q == ')') {
 		while((*--q != '('));
-		*q = (char)NULL;
+		*q = 0;
 	}
 	(void)strip_post_ws(p);
 
@@ -363,8 +377,14 @@ bool_t standardise(char *str, bool_t *linestart)
 	*linestart = False;
 
 	if((p = strchr(str, '\n'))) {
-		*p = (char)NULL;
+		*p = 0;
 		*linestart = True;
+
+		/* If the line ended in "\r\n", then drop the '\r' too */
+		sl = strlen(str);
+		if(sl >= 1 && str[sl - 1] == '\r') {
+			str[sl - 1] = 0;
+		}
 	}
 	return(leadingdot);
 }
@@ -384,7 +404,7 @@ void revaliases(struct passwd *pw)
 		while(fgets(buf, sizeof(buf), fp)) {
 			/* Make comments invisible */
 			if((p = strchr(buf, '#'))) {
-				*p = (char)NULL;
+				*p = 0;
 			}
 
 			/* Ignore malformed lines and comments */
@@ -519,7 +539,7 @@ void rcpt_save(char *str)
 #endif
 
 	/* Ignore missing usernames */
-	if(*str == (char)NULL) {
+	if(*str == 0) {
 		return;
 	}
 
@@ -576,7 +596,7 @@ void rcpt_parse(char *str)
 		}
 
 		/* End of string? */
-		if(*(q + 1) == (char)NULL) {
+		if(*(q + 1) == 0) {
 			got_addr = True;
 		}
 
@@ -584,7 +604,7 @@ void rcpt_parse(char *str)
 		if((*q == ',') && (in_quotes == False)) {
 			got_addr = True;
 
-			*q = (char)NULL;
+			*q = 0;
 		}
 
 		if(got_addr) {
@@ -610,30 +630,30 @@ int crammd5(char *challengeb64, char *username, char *password, char *responseb6
 	unsigned char digascii[MD5_DIGEST_LEN * 2];
 	unsigned char challenge[(BUF_SZ + 1)];
 	unsigned char response[(BUF_SZ + 1)];
-	unsigned char secret[(MD5_BLOCK_LEN + 1)]; 
+	unsigned char secret[(MD5_BLOCK_LEN + 1)];
 
 	memset (secret,0,sizeof(secret));
 	memset (challenge,0,sizeof(challenge));
-	strncpy (secret, password, sizeof(secret));	
+	strncpy ((char *)secret, password, sizeof(secret) - 1);
 	if (!challengeb64 || strlen(challengeb64) > sizeof(challenge) * 3 / 4)
 		return 0;
-	from64tobits(challenge, challengeb64);
+	from64tobits((char *)challenge, challengeb64);
 
-	hmac_md5(challenge, strlen(challenge), secret, strlen(secret), digest);
+	hmac_md5(challenge, strlen((char *)challenge), secret, strlen((char *)secret), digest);
 
 	for (i = 0; i < MD5_DIGEST_LEN; i++) {
 		digascii[2 * i] = hextab[digest[i] >> 4];
 		digascii[2 * i + 1] = hextab[(digest[i] & 0x0F)];
 	}
-	digascii[MD5_DIGEST_LEN * 2] = '\0';
+	digascii[MD5_DIGEST_LEN * 2 - 1] = '\0';
 
 	if (sizeof(response) <= strlen(username) + sizeof(digascii))
 		return 0;
 	
-	strncpy (response, username, sizeof(response) - sizeof(digascii) - 2);
-	strcat (response, " ");
-	strcat (response, digascii);
-	to64frombits(responseb64, response, strlen(response));
+	strncpy ((char *)response, username, sizeof(response) - sizeof(digascii) - 2);
+	strcat ((char *)response, " ");
+	strcat ((char *)response, (char *)digascii);
+	to64frombits((unsigned char *)responseb64, response, strlen((char *)response));
 
 	return 1;
 }
@@ -676,7 +696,7 @@ void header_save(char *str)
 	if(strncasecmp(ht->string, "From:", 5) == 0) {
 #if 1
 		/* Hack check for NULL From: line */
-		if(*(p + 6) == (char)NULL) {
+		if(*(p + 6) == 0) {
 			return;
 		}
 #endif
@@ -685,9 +705,6 @@ void header_save(char *str)
 		if(override_from == True) {
 			uad = from_strip(ht->string);
 		}
-		else {
-			return;
-		}
 #endif
 		have_from = True;
 	}
@@ -741,7 +758,7 @@ void header_parse(FILE *stream)
 	size_t size = BUF_SZ, len = 0;
 	char *p = (char *)NULL, *q;
 	bool_t in_header = True;
-	char l = (char)NULL;
+	char l = 0;
 	int c;
 
 	while(in_header && ((c = fgetc(stream)) != EOF)) {
@@ -758,6 +775,14 @@ void header_parse(FILE *stream)
 		}
 		len++;
 
+		if(l == '\r' && c == '\n') {
+			/* Properly handle input that already has "\r\n"
+			   line endings; see https://bugs.debian.org/584162 */
+			l = (len >= 2 ? *(q - 2) : '\n');
+			q--;
+			len--;
+		}
+
 		if(l == '\n') {
 			switch(c) {
 				case ' ':
@@ -776,11 +801,13 @@ void header_parse(FILE *stream)
 						in_header = False;
 
 				default:
-						*q = (char)NULL;
+						*q = 0;
 						if((q = strrchr(p, '\n'))) {
-							*q = (char)NULL;
+							*q = 0;
+						}
+						if(len > 0) {
+							header_save(p);
 						}
-						header_save(p);
 
 						q = p;
 						len = 0;
@@ -790,35 +817,12 @@ void header_parse(FILE *stream)
 
 		l = c;
 	}
-	if(in_header) {
-		if(l == '\n') {
-			switch(c) {
-				case ' ':
-				case '\t':
-						/* Must insert '\r' before '\n's embedded in header
-						   fields otherwise qmail won't accept our mail
-						   because a bare '\n' violates some RFC */
-						
-						*(q - 1) = '\r';	/* Replace previous \n with \r */
-						*q++ = '\n';		/* Insert \n */
-						len++;
-						
-						break;
-
-				case '\n':
-						in_header = False;
-
-				default:
-						*q = (char)NULL;
-						if((q = strrchr(p, '\n'))) {
-							*q = (char)NULL;
-						}
-						header_save(p);
-
-						q = p;
-						len = 0;
-			}
+	if(in_header && l == '\n') {
+		/* Got EOF while reading the header */
+		if((q = strrchr(p, '\n'))) {
+			*q = 0;
 		}
+		header_save(p);
 	}
 	(void)free(p);
 }
@@ -876,7 +880,7 @@ bool_t read_config()
 		char *rightside;
 		/* Make comments invisible */
 		if((p = strchr(buf, '#'))) {
-			*p = (char)NULL;
+			*p = 0;
 		}
 
 		/* Ignore malformed lines and comments */
@@ -1107,11 +1111,11 @@ int smtp_open(char *host, int port)
 #ifdef INET6
 	struct addrinfo hints, *ai0, *ai;
 	char servname[NI_MAXSERV];
-	int s;
+	int s = 0;
 #else
 	struct sockaddr_in name;
 	struct hostent *hent;
-	int i, s, namelen;
+	int i, s = 0, namelen;
 #endif
 
 #ifdef HAVE_SSL
@@ -1120,7 +1124,7 @@ int smtp_open(char *host, int port)
 
 	/* Init SSL stuff */
 	SSL_CTX *ctx;
-	SSL_METHOD *meth;
+	const SSL_METHOD *meth;
 	X509 *server_cert;
 
 	SSL_load_error_strings();
@@ -1286,8 +1290,12 @@ fd_getc() -- Read a character from an fd
 ssize_t fd_getc(int fd, void *c)
 {
 #ifdef HAVE_SSL
-	if(use_tls == True) { 
-		return(SSL_read(ssl, c, 1));
+	if(use_tls == True) {
+		int attempt = 3;
+		int ret = 0;
+		while (attempt-- > 0 && ret == 0)
+			ret = SSL_read(ssl, c, 1);
+		return ret;
 	}
 #endif
 	return(read(fd, c, 1));
@@ -1310,7 +1318,7 @@ char *fd_gets(char *buf, int size, int fd)
 			buf[i++] = c;
 		}
 	}
-	buf[i] = (char)NULL;
+	buf[i] = 0;
 
 	return(buf);
 }
@@ -1409,6 +1417,7 @@ ssmtp() -- send the message (exactly one) from stdin to the mailhub SMTP port
 int ssmtp(char *argv[])
 {
 	char b[(BUF_SZ + 2)], *buf = b+1, *p, *q;
+	char *remote_addr;
 #ifdef MD5AUTH
 	char challenge[(BUF_SZ + 1)];
 #endif
@@ -1457,7 +1466,7 @@ int ssmtp(char *argv[])
 	from = from_format(uad, override_from);
 
 	/* Now to the delivery of the message */
-	(void)signal(SIGALRM, (void(*)())handler);	/* Catch SIGALRM */
+	(void)signal(SIGALRM, (sighandler_t)handler);	/* Catch SIGALRM */
 	(void)alarm((unsigned) MAXWAIT);			/* Set initial timer */
 	if(setjmp(TimeoutJmpBuf) != 0) {
 		/* Then the timer has gone off and we bail out */
@@ -1500,7 +1509,7 @@ int ssmtp(char *argv[])
 			if(smtp_read(sock, buf) != 3) {
 				die("Server rejected AUTH CRAM-MD5 (%s)", buf);
 			}
-			strncpy(challenge, strchr(buf,' ') + 1, sizeof(challenge));
+			strncpy(challenge, strchr(buf,' ') + 1, sizeof(challenge) - 1);
 
 			memset(buf, 0, bufsize);
 			crammd5(challenge, auth_user, auth_pass, buf);
@@ -1508,7 +1517,7 @@ int ssmtp(char *argv[])
 		else {
 #endif
 		memset(buf, 0, bufsize);
-		to64frombits(buf, auth_user, strlen(auth_user));
+		to64frombits((unsigned char *)buf, (unsigned char *)auth_user, strlen(auth_user));
 		if (use_oldauth) {
 			outbytes += smtp_write(sock, "AUTH LOGIN %s", buf);
 		}
@@ -1520,7 +1529,7 @@ int ssmtp(char *argv[])
 			}
 			/* we assume server asked us for Username */
 			memset(buf, 0, bufsize);
-			to64frombits(buf, auth_user, strlen(auth_user));
+			to64frombits((unsigned char *)buf, (unsigned char *)auth_user, strlen(auth_user));
 			outbytes += smtp_write(sock, buf);
 		}
 
@@ -1530,7 +1539,7 @@ int ssmtp(char *argv[])
 		}
 		memset(buf, 0, bufsize);
 
-		to64frombits(buf, auth_pass, strlen(auth_pass));
+		to64frombits((unsigned char *)buf, (unsigned char *)auth_pass, strlen(auth_pass));
 #ifdef MD5AUTH
 		}
 #endif
@@ -1612,6 +1621,10 @@ int ssmtp(char *argv[])
 		outbytes += smtp_write(sock, "From: %s", from);
 	}
 
+	if((remote_addr=getenv("REMOTE_ADDR"))) {
+		outbytes += smtp_write(sock, "X-Originating-IP: %s", remote_addr);
+	}
+
 	if(have_date == False) {
 		outbytes += smtp_write(sock, "Date: %s", arpadate);
 	}
@@ -1657,12 +1670,12 @@ int ssmtp(char *argv[])
 			outbytes += smtp_write(sock, "%s", leadingdot ? b : buf);
 		} else {
 			if (log_level > 0) {
-				log_event(LOG_INFO, "Sent a very long line in chunks");
+				log_event(LOG_INFO, "Sending a partial line");
 			}
 			if (leadingdot) {
-				outbytes += fd_puts(sock, b, sizeof(b));
+				outbytes += fd_puts(sock, b, strlen(b));
 			} else {
-				outbytes += fd_puts(sock, buf, bufsize);
+				outbytes += fd_puts(sock, buf, strlen(buf));
 			}
 		}
 		(void)alarm((unsigned) MEDWAIT);
@@ -1737,7 +1750,7 @@ char **parse_options(int argc, char *argv[])
 		j = 0;
 
 		add = 1;
-		while(argv[i][++j] != (char)NULL) {
+		while(argv[i][++j] != 0) {
 			switch(argv[i][j]) {
 #ifdef INET6
 			case '6':
