@@ -1,7 +1,5 @@
 /*
 
- $Id: ssmtp.h,v 2.7 2002/09/27 13:18:13 matt Exp $
-
  See COPYRIGHT for the license
 
 */
@@ -22,6 +20,12 @@
 #define MAXARGS  _POSIX_ARG_MAX
 #endif
 
+/* ssmtp assumes MAXHOSTNAMELEN is alwyas in sys/param.h this is
+   not always the case (System V/Solaris) */
+#ifndef MAXHOSTNAMELEN
+#define MAXHOSTNAMELEN 256
+#endif
+
 typedef enum {False, True} bool_t;
 
 struct string_list {
